# -*- coding: utf-8 -*-
from odoo import api, models, fields, _
from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.addons.saas_worker.const import TEST, SAAS_BASE_URL, SCHEME
from odoo.exceptions import AccessError, UserError

from .sale_subscription import SalesRules

from dateutil.relativedelta import relativedelta

import json
import requests

import logging
_logger = logging.getLogger(__name__)

WHITE_LISTED_FIELDS = ['internal_note', 'internal_note_display']

class SalesOrder(models.Model):
    _inherit = 'sale.order'

    # add tracking to referrer_id
    referrer_id = fields.Many2one('res.partner', 'Referrer', domain=[('grade_id', '!=', False)], tracking=True)
    enterprise_intercompany_invoice = fields.Boolean('Intercompany Invoicing',
                                                     compute="_need_intercompany_invoice")
    partner_status = fields.Html(
        help='The partner has a level and a valid partnership subscription',
        compute='_compute_partner_status'
    )

    def _check_so(self):
        self.ensure_one()
        SalesRules()._check_so(self)

    def _check_sub(self):
        self.ensure_one()
        SalesRules()._check_sub(self)

    @api.model_create_multi
    def create(self, vals_list):
        res = super().create(vals_list)

        if not self.env.su:
            for so in res:
                if so.is_subscription:
                    so._check_sub()
                else:
                    so._check_so()

        return res


    @api.multi
    def write(self, vals):

        # Prevent write on next_invoice_date if the SO is locked
        if any(so.state == 'sale' and so.locked for so in self) and 'next_invoice_date' in vals:
            raise UserError(_("You cannot modify the next invoice date of a locked Sales Order."))

        res = super().write(vals)

        if not self.env.su and any(field not in WHITE_LISTED_FIELDS for field in vals):
            for so in self:
                if so.is_subscription:
                    so._check_sub()
                else:
                    so._check_so()

        return res

    # Requested by CHE: forbid confirming/sending a Sales Order when the selected
    # customer has no country selected yet. This is important for proper
    # assignation of fiscal positions and taxes, and ultimately, final price.
    def _check_partner_country(self):
        if not self.env.su:
            # do not raise while in sudo() mode (ex. when processing payment notifications)
            for order in self:
                if order.state in ('draft', 'sent') and not order.partner_id.country_id:
                    raise AccessError(
                        "Please first make sure the customer's country is set, and then "
                        "double-check the fiscal position and taxes, if relevant.\n"
                        "Thanks a lot! :-)")

    def _ping_synchronize_metabase(self):
        if not self.env.context.get('no_ping'):
            databases = self.database_ids.filtered(lambda db: db.hosting in ('saas',))
            if len(databases) == 1:
                db = databases[0]
                try:
                    requests.post(
                        '%s://%s.%s/saas_worker/synchronize_metabase' % (SCHEME, db.db_name, SAAS_BASE_URL),
                        data={'dbuuids': json.dumps([db.db_uuid])},
                        timeout=(3.05, 0.001),  # (connectionTimeout, readTimeout)
                        auth=SaasAuth(),
                    )
                except requests.exceptions.ReadTimeout:
                    pass
                except Exception as e:
                    _logger.info('Ping failed: %s', e)

    def action_quotation_send(self):
        self._check_partner_country()
        return super().action_quotation_send()

    def action_quotation_sent(self):
        self._check_partner_country()
        return super().action_quotation_sent()

    # TBE, still the same fix, this time for the prices
    def _cart_update(self, product_id=None, line_id=None, add_qty=0, set_qty=0, **kwargs):
        product = self.env['product.product'].browse(product_id).exists()
        if product.recurring_invoice and (add_qty or set_qty > 0):
            # call to _precommit_check_order_quotation_template bound to this record
            # to ensure that after adding a recurring product we correctly set
            # a quotation template on the website order (cart)
            self.env.cr.precommit.add(self._precommit_check_order_quotation_template)

        res = super()._cart_update(product_id=product_id, line_id=line_id,
            add_qty=add_qty, set_qty=set_qty, **kwargs)

        if product and (product.is_odoo_app or product.is_odoo_user):
            order_line = self.env['sale.order.line'].sudo().browse(res['line_id'])
            if order_line.exists():
                order_line._compute_price_unit()
                order_line._compute_amount()
        return res

    def _precommit_check_order_quotation_template(self):
        """Ensure that a quotation template is set for recurring order (cart)."""
        if self.plan_id and not self.sale_order_template_id:
            cart_products = self.order_line.product_id
            templates = self.env["sale.order.template"].search([('plan_id', '=', self.plan_id.id)]).filtered(
                lambda x: not (cart_products - x.sale_order_template_line_ids.product_id)
            )
            if len(templates) == 1:
                # we autodected the the missing quotation template, set it back
                self.sale_order_template_id = templates[0]
            else:
                # if we are unable to compute the exact quotation template
                # simply prevent the cart update to proceed.
                raise UserError("Unable to update your cart, please refresh your page")

    def _need_intercompany_invoice(self):
        """Check if a foreign quote requires invoicing in Odoo S.A."""
        Acquirer = self.env['payment.provider'].sudo()
        ogone_acquirers = Acquirer.search([('code', '=', 'ogone'),
                                           ('state', '=', 'test' if TEST else 'enabled')])
        for order in self:
            payment_tx_id = order.transaction_ids.filtered(lambda x: x.state in ['authorized', 'done'])[:1]
            order.enterprise_intercompany_invoice = False
            if order.company_id in (
                    self.env.ref('openerp_enterprise.company_dmcc'),
                    self.env.ref('openerp_enterprise.company_ke'),
            ):
                continue
            if payment_tx_id and payment_tx_id.provider_id in ogone_acquirers and \
               order.company_id not in ogone_acquirers.mapped('company_id'):
                order.enterprise_intercompany_invoice = True

    def action_confirm(self):
        self._check_partner_country()
        if self.env.context.get('check_existing_sub') and self.is_subscription:
            today = fields.Date.context_today(self)
            domain = [
                ('subscription_state', 'in', ['3_progress', '4_paused', '6_churn']),
                '|',
                ('end_date', '>', today - relativedelta(months=6)),
                ('end_date', '=', False),
                ('id', 'not in', [self.id] + self.subscription_id.ids),
            ]
            existing_subscription = self._check_duplicate(self.commercial_partner_id.email or self.partner_id.email, domain)[:1]
            if existing_subscription:
                return {
                    'type': 'ir.actions.act_window',
                    'view_mode': 'form',
                    'res_model': 'openerp.enterprise.existing_subscription',
                    'target': 'new',
                    'context': dict(
                        self.env.context,
                        default_new_subscription_id=self.id,
                        default_existing_subscription_id=existing_subscription.id,
                    ),
                }

        res = super().action_confirm()

        # Order the Odoo mugs if there are any
        self._order_confirm_deliver_odoo_mug()
        # force sync (after confirmation), if not in batch
        if len(self) == 1 and self.is_subscription:
            self._cr.postcommit.add(self._ping_synchronize_metabase)

        template = self.env.ref('openerp_enterprise.email_subscription_open')
        for order in self:
            if order.is_subscription and not order.subscription_id and order._is_enterprise_plans():
                order.message_post_with_source(template)

        return res

    def prepare_renewal_order(self):
        if self.env.context.get('check_existing_sub') and self.is_subscription:
            if self.start_date == self.next_invoice_date:
                return {
                    'type': 'ir.actions.act_window',
                    'view_mode': 'form',
                    'res_model': 'openerp.enterprise.existing_subscription',
                    'target': 'new',
                    'context': dict(
                        self.env.context, default_new_subscription_id=self.id,
                        default_existing_subscription_id=self.subscription_id.id,
                        default_state='renew',
                    ),
                }
        res = super().prepare_renewal_order()

        return res

    def _confirm_subscription(self):
        res = super()._confirm_subscription()
        # Set SaaS values on subscription and invoice in Odoo S.A. when needed.
        #
        # - Copy the integrating partner field
        # - Set the "Self Odoo"/"Sold Odoo" tag
        for s in self:
            # assign the self/not self tag on the subscription for easier classifcation
            oe_self_tag = self.env.ref('openerp_enterprise.self_subscription_crm_tag')
            if (
                oe_self_tag in s.tag_ids
                or (s.origin and 'OE Self' in s.origin.split(';'))  # keep the old code for retro compatibility
            ):
                s.subscription_tag_ids |= self.env.ref('openerp_enterprise.self_subscription_tag')
            else:
                s.subscription_tag_ids |= self.env.ref('openerp_enterprise.sold_subscription_tag')
        return res

    def _order_confirm_deliver_odoo_mug(self):
        """ This method is called when a sale order is confirmed,
            It will check the order lines for the Odoo Mug products
            and call the Gooten API to make the order
        """
        product_odoo_mug = self.env.ref('openerp_enterprise.product_odoo_merch_mug')

        for order in self.filtered(lambda x: x.state == 'sale'):
            gooten_order_ids = []
            mug_lines = order.order_line.filtered(lambda x: x.product_id == product_odoo_mug)

            for line in mug_lines:

                name = order.partner_shipping_id.display_name.split(' ')
                for i in range(int(line.product_uom_qty)):
                    res = self.env['openerp.enterprise.database.user']._gooten_order_mug(
                        firstName=' '.join(name[:1] or name),
                        lastName=' '.join(name[1:] or name),
                        addrLine1=order.partner_shipping_id.street,
                        addrLine2=order.partner_shipping_id.street2 or '',
                        addrCity=order.partner_shipping_id.city,
                        addrState=order.partner_shipping_id.state_id.name or '',
                        addrPostalCode=order.partner_shipping_id.zip,
                        addrCountryCode=order.partner_shipping_id.country_id.code,
                        phone=order.partner_shipping_id.phone,
                        email=order.partner_shipping_id.email,
                    )
                    if res:
                        gooten_order_ids.append(res)

            if mug_lines:
                # print on the sale order, the gooten order ids
                if gooten_order_ids:
                    order.message_post(body=_('Ordered %d mugs to Gooten API, with order ids [%s].') % (len(gooten_order_ids), ', '.join(gooten_order_ids)))
                else:
                    order.message_post(body=_('Error while ordering mugs to Gooten API.'))

    @api.depends('partner_id.grade_id', 'partner_id.enterprise_subscription_ids')
    def _compute_partner_status(self):
        for sub in self:
            p = sub.referrer_id
            is_ok = p.grade_id and p.enterprise_subscription_ids.filtered(lambda s: s.enterprise_is_partnership)
            sub.partner_status = is_ok and "&#128994;" or "&#128308;"

    def _prepare_invoice(self):
        # avoid conflict of check_company during payment reconciliation
        # cron 'Fix Partner Companies' will do it later btw.
        if self.partner_id.company_id:
            self.partner_id.company_id = False
        return super()._prepare_invoice()

    def _portal_switch_yearly_url(self):
        self.ensure_one()
        action = self.prepare_renewal_order()

        new_so = self.browse(action['res_id'])

        if new_so._is_standard_plan():
            template = self.env.ref('openerp_enterprise.standard_template_odoo_yearly')
        else:
            template = self.env.ref('openerp_enterprise.custom_template_odoo_yearly')

        new_so.write({
            'sale_order_template_id': template.id,
            'plan_id': template.plan_id.id,
            'state': 'sent',
        })
        new_so.order_line._compute_product_uom()
        discount_product = (
            self.env.ref('openerp_enterprise.product_odoo_user_standard_discount') |
            self.env.ref('openerp_enterprise.product_odoo_user_custom_discount')
        )
        discount_line = new_so.order_line.filtered(lambda l: l.product_id in discount_product)
        if discount_line:
            discount_line.unlink()

        return new_so.get_portal_url()

    def _can_switch_yearly(self):
        self.ensure_one()
        monthly_pack = self.env.ref("openerp_enterprise.monthly_success_pack_template")
        if self.sale_order_template_id == monthly_pack:
            return False
        if self.plan_id.billing_period_unit != 'month':
            return False
        if self.start_date == self.next_invoice_date:
            return False
        if self.subscription_state != '3_progress':
            return False
        return True

    def _get_switch_url(self):
        self.ensure_one()
        return "/my/subscription/switch-yearly/{}/{}".format(self.id, self._portal_ensure_token())

    def action_open_discount_wizard(self):
        self.ensure_one()
        res = super().action_open_discount_wizard()

        if self.plan_id.billing_period_unit != 'year':
            return res

        discount = {1: 0.00, 2: 0.05, 3: 0.10, 4: 0.10, 5: 0.15, }.get(self.plan_id.billing_period_value, 0)
        wizard = self.env['sale.order.discount'].create({
            'sale_order_id': self.id,
            'discount_type': 'so_discount',
            'discount_percentage': discount,
        })
        res['res_id'] = wizard.id
        return res


class SalesOrderDiscount(models.TransientModel):
    _inherit = 'sale.order.discount'

    def _get_discount_product(self):
        if self.sale_order_id.plan_id.billing_period_unit == 'year':
            return self.env.ref('openerp_enterprise.product_multiyear_discount')
        else:
            return super()._get_discount_product()

    def action_apply_discount(self):
        if self.discount_type == 'so_discount' and self.discount_percentage > 0.20:
            raise UserError(_("You cannot apply more than 20% as discount."))
        return super().action_apply_discount()

    def _prepare_discount_line_values(self, product, amount, taxes, description=None):

        odoo_sub_categ = self.env.ref('openerp_enterprise.product_category_odoo_users')
        discountable_lines_for_current_tax = self.sale_order_id.order_line.filtered(
            lambda l: l.product_id.categ_id == odoo_sub_categ and l.tax_id == taxes
        )

        amount = sum(discountable_lines_for_current_tax.mapped('price_subtotal')) * self.discount_percentage
        return super()._prepare_discount_line_values(product, amount, taxes, description=description)
